from __future__ import annotations

import dataclasses
import enum
from typing import TYPE_CHECKING, Any, Literal, Protocol, overload

from typing_extensions import Self

if TYPE_CHECKING:
    # Solve TypeError: Cannot create a consistent method resolution
    class HasFieldName(Protocol):
        field_name: str
else:

    class HasFieldName:
        field_name: str


class Expression:
    """
    Base class for all expressions.
    """

    def __invert__(self) -> Expression:
        """
        ~self
        """
        match self:
            case NotExpression(expr):
                return expr
            case CombineExpression("AND", left, right):
                return CombineExpression("OR", ~left, ~right)
            case CombineExpression("OR", left, right):
                return CombineExpression("AND", ~left, ~right)
            case CompareExpression():
                return NotExpression(self)
            case _:
                raise NotImplementedError

    @overload
    def __and__(self, other: None) -> Self: ...

    @overload
    def __and__(self, other: Expression) -> CombineExpression: ...

    def __and__(self, other):
        """
        self & other
        """
        if other is None:
            return self
        if not isinstance(other, Expression):
            return NotImplemented
        return CombineExpression("AND", self, other)

    @overload
    def __rand__(self, other: None) -> Self: ...

    @overload
    def __rand__(self, other: Expression) -> CombineExpression: ...

    def __rand__(self, other):
        """
        other & self
        """
        if other is None:
            return self
        if not isinstance(other, Expression):
            return NotImplemented
        return CombineExpression("AND", other, self)

    @overload
    def __or__(self, other: None) -> Self: ...

    @overload
    def __or__(self, other: Expression) -> CombineExpression: ...

    def __or__(self, other):
        """
        self | other
        """
        if other is None:
            return self
        if not isinstance(other, Expression):
            return NotImplemented
        return CombineExpression("OR", self, other)

    @overload
    def __ror__(self, other: None) -> Self: ...

    @overload
    def __ror__(self, other: Expression) -> CombineExpression: ...

    def __ror__(self, other):
        """
        other | self
        """
        if other is None:
            return self
        if not isinstance(other, Expression):
            return NotImplemented
        return CombineExpression("OR", other, self)

    def compile(self) -> dict[str, Any]:
        return compile_expression(self)


@dataclasses.dataclass
class RawExpression(Expression):
    raw: dict[str, Any]


class CompareMixin(HasFieldName):
    def __eq__(self, other: Any) -> CompareExpression:
        """
        self == other
        """
        return CompareExpression(self, "==", other)

    def __ne__(self, other: Any) -> CompareExpression:
        """
        self != other
        """
        return CompareExpression(self, "!=", other)

    def __lt__(self, other: Any) -> CompareExpression:
        """
        self < other
        """
        return CompareExpression(self, "<", other)

    def __le__(self, other: Any) -> CompareExpression:
        """
        self <= other
        """
        return CompareExpression(self, "<=", other)

    def __gt__(self, other: Any) -> CompareExpression:
        """
        self > other
        """
        return CompareExpression(self, ">", other)

    def __ge__(self, other: Any) -> CompareExpression:
        """
        self >= other
        """
        return CompareExpression(self, ">=", other)


@dataclasses.dataclass(repr=False)
class BetterReprMixin:
    def __repr__(self) -> str:
        names = self.__dataclass_fields__.keys()

        attrs_str = []
        for name in names:
            v = getattr(self, name)
            if hasattr(v, "field_name"):
                repr_str = getattr(v, "field_name")
            else:
                repr_str = repr(v)
            attrs_str.append(f"{name}={repr_str}")
        return f"{self.__class__.__name__}({', '.join(attrs_str)})"


@dataclasses.dataclass
class CombineExpression(Expression):
    operator: Literal["AND", "OR"]
    left: Expression
    right: Expression


@dataclasses.dataclass(repr=False)
class CompareExpression(Expression, BetterReprMixin):
    field: HasFieldName
    operator: Literal[">=", "<=", ">", "<", "==", "!="]
    arg: Any


@dataclasses.dataclass
class NotExpression(Expression):
    expr: CompareExpression


def compile_expression(expr: Expression) -> dict[str, Any]:
    match expr:
        case RawExpression(raw):
            return raw
        case CompareExpression(field, "==", arg):
            return {
                field.field_name: arg if not isinstance(arg, enum.Enum) else arg.value
            }
        case CompareExpression(field, "!=", arg):
            return {
                field.field_name: {
                    "$ne": arg if not isinstance(arg, enum.Enum) else arg.value
                }
            }
        case CompareExpression(field, ">", arg):
            return {
                field.field_name: {
                    "$gt": arg if not isinstance(arg, enum.Enum) else arg.value
                }
            }
        case CompareExpression(field, ">=", arg):
            return {
                field.field_name: {
                    "$gte": arg if not isinstance(arg, enum.Enum) else arg.value
                }
            }
        case CompareExpression(field, "<", arg):
            return {
                field.field_name: {
                    "$lt": arg if not isinstance(arg, enum.Enum) else arg.value
                }
            }
        case CompareExpression(field, "<=", arg):
            return {
                field.field_name: {
                    "$lte": arg if not isinstance(arg, enum.Enum) else arg.value
                }
            }
        case CombineExpression("AND", left, right):
            return {"$and": [compile_expression(left), compile_expression(right)]}
        case CombineExpression("OR", left, right):
            return {"$or": [compile_expression(left), compile_expression(right)]}
        case NotExpression(inner):
            return {
                inner.field.field_name: {
                    "$not": compile_expression(inner).pop(inner.field.field_name)
                }
            }
        case _:
            raise NotImplementedError(f"Unsupported expression: {expr}")


class OrderByMixin(HasFieldName):
    def __neg__(self) -> OrderBy:
        """
        -self
        """
        return OrderBy(self, -1)

    def __pos__(self) -> OrderBy:
        """
        +self
        """
        return OrderBy(self, +1)


@dataclasses.dataclass(repr=False)
class OrderBy(BetterReprMixin):
    field: HasFieldName
    order: Literal[-1, +1]



"""
Synchronize the asyncio module to the root module.
"""

import pathlib


def main(just_check: bool = False):
    exit_code = 0
    sync_dir = pathlib.Path(__file__).absolute().parent
    asyncio_dir = pathlib.Path(__file__).absolute().parent / "asyncio"
    for path in asyncio_dir.glob("*.py"):
        new_file_path = sync_dir / path.name
        content = path.read_text()
        content = (
            content.replace(
                "from pymongo.asynchronous.mongo_client import AsyncMongoClient as MongoClient",
                "from pymongo.synchronous.mongo_client import MongoClient",
            )
            .replace("typedmongo.asyncio", "typedmongo")
            .replace("async def ", "def ")
            .replace("await ", "")
            .replace("async for ", "for ")
            .replace("async with ", "with ")
            .replace("AsyncIterable", "Iterable")
        )
        if just_check:
            if content != new_file_path.read_text():
                print(f"File {new_file_path} is not synchronized.")
                exit_code = 1
        else:
            new_file_path.write_text(content)
    return exit_code


if __name__ == "__main__":
    import sys

    sys.exit(main("--check" in " ".join(sys.argv[1:])))



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, List, get_args

from bson import ObjectId
from bson.errors import InvalidId
from marshmallow import ValidationError, fields

if TYPE_CHECKING:
    from .fields import Field


class MarshamallowObjectId(fields.Field):
    def _serialize(self, value: ObjectId, attr: str | None, obj: Any, **kwargs) -> str:
        return str(value)

    def _deserialize(
        self, value: str | ObjectId | bytes, attr: str | None, data: Any, **kwargs
    ) -> ObjectId:
        try:
            return ObjectId(value)
        except (InvalidId, TypeError):
            raise ValidationError("Invalid ObjectId.")


class MarshamallowDateTime(fields.DateTime):
    def _deserialize(
        self, value: str | datetime, attr: str | None, data: Any, **kwargs
    ) -> datetime:
        if isinstance(value, datetime):
            return value
        return super()._deserialize(value, attr, data, **kwargs)


class MarshamallowLiteral(fields.Field):
    def __init__(self, literal: Any, **kwargs):
        self.enum = get_args(literal)
        super().__init__(**kwargs)

    def _deserialize(self, value: Any, attr: str | None, data: Any, **kwargs) -> Any:
        if value not in self.enum:
            raise ValidationError(f"Value must be {self.enum}")
        return value


class MarshamallowUnion(fields.Field):
    def __init__(self, fields: List[Field[Any]], **kwargs: Any):
        self._candidate_fields = fields
        self._candidate_fields_map = {field.field_type: field for field in fields}
        super().__init__(**kwargs)

    def _serialize(self, value: Any, attr: str | None, obj: Any, **kwargs):
        candidate_field = self._candidate_fields_map.get(type(value))
        if candidate_field:
            return candidate_field.dump(value)

        raise TypeError(
            f"Unable to serialize value {value} with any of the candidate fields"
        )

    def _deserialize(self, value: Any, attr: str, data: Any, **kwargs: Any):
        errors = []
        try_field = self._candidate_fields_map.get(type(value))
        if try_field:
            return try_field.load(
                try_field.marshamallow.deserialize(value, attr, data, **kwargs)
            )

        for candidate_field in self._candidate_fields:
            try:
                return candidate_field.load(
                    candidate_field.marshamallow.deserialize(
                        value, attr, data, **kwargs
                    )
                )
            except ValidationError as exc:
                errors.append(exc.messages)
        raise ValidationError(message=errors, field_name=attr)



from __future__ import annotations

import datetime
import enum
import uuid
from decimal import Decimal
from typing import Literal

import pytest
from bson import ObjectId

import typedmongo as mongo
from typedmongo.fields import type_to_field
from typedmongo.expressions import Expression


class MongoDocument(mongo.Document):
    __abstract__ = True

    _id: mongo.StringField = mongo.StringField(default=lambda: uuid.uuid4().hex)


class Wallet(mongo.Document):
    balance: mongo.DecimalField
    created_at: mongo.DateTimeField = mongo.DateTimeField(
        default=lambda: datetime.datetime.now(datetime.timezone.utc)
    )


class Social(mongo.Document):
    site: mongo.StringField
    user: mongo.StringField
    updated_at: mongo.DateTimeField = mongo.DateTimeField(
        default=lambda: datetime.datetime.now(datetime.timezone.utc)
    )


class Place(enum.Enum):
    ONE = 1
    TWO = 2
    THREE = 3


class User(MongoDocument):
    name: mongo.StringField
    gender: mongo.LiteralField[Literal["m", "f"]]
    age: mongo.IntegerField
    place: mongo.EnumField[Place] = mongo.EnumField(Place, default=Place.ONE)
    tags: mongo.ListField[str]
    wallet: mongo.EmbeddedField[Wallet]
    created_at: mongo.DateTimeField = mongo.DateTimeField(
        default=lambda: datetime.datetime.now(datetime.timezone.utc)
    )
    children: mongo.ListField[User]
    socials: mongo.ListField[Social] = mongo.ListField(
        mongo.EmbeddedField(Social), default=list
    )
    extra: mongo.DictField = mongo.DictField(default=dict)


User.__lazy_init_fields__()


@pytest.mark.parametrize(
    "expression, repr_str",
    [
        (
            User.name == "Aber",
            "CompareExpression(field=name, operator='==', arg='Aber')",
        ),
        (
            User.age >= 18,
            "CompareExpression(field=age, operator='>=', arg=18)",
        ),
        (
            User.tags == "a",
            "CompareExpression(field=tags, operator='==', arg='a')",
        ),
        (
            User.tags[0] == "0",
            "CompareExpression(field=tags.0, operator='==', arg='0')",
        ),
        (
            User.wallet._.balance > 1,
            "CompareExpression(field=wallet.balance, operator='>', arg=1)",
        ),
        (
            User.children[0].name == "Yue",
            "CompareExpression(field=children.0.name, operator='==', arg='Yue')",
        ),
        (
            User.children._.age >= 18,
            "CompareExpression(field=children.age, operator='>=', arg=18)",
        ),
        (
            User.extra == {"a": "b"},
            "CompareExpression(field=extra, operator='==', arg={'a': 'b'})",
        ),
    ],
)
def test_expression(expression, repr_str):
    assert isinstance(expression, Expression)
    assert repr(expression) == repr_str


def test_list_field():
    now = datetime.datetime.now(datetime.timezone.utc)
    user = User.load(
        {
            "name": "Aber",
            "gender": "m",
            "age": 18,
            "tags": ["a", "b"],
            "wallet": {"balance": 100},
            "children": [],
            "socials": [
                {"site": "github.com", "user": "abersheeran", "updated_at": now}
            ],
        }
    )
    assert user.socials == [
        Social(site="github.com", user="abersheeran", updated_at=now)
    ]
    assert isinstance(user.socials[0].updated_at, datetime.datetime)
    assert (
        user.dump()["socials"][0]["updated_at"]
        == user.socials[0].updated_at.isoformat()
    )


def test_field_default():
    user = User.load(
        {
            "name": "Aber",
            "gender": "m",
            "age": 18,
            "tags": ["a", "b"],
            "wallet": {"balance": 100},
            "children": [],
        }
    )
    assert isinstance(user._id, str)
    assert isinstance(user.created_at, datetime.datetime)
    assert user.place == Place.ONE

    user = User.load(
        {
            "name": "Aber",
            "age": 18,
            "tags": ["a", "b"],
            "wallet": {"balance": 100},
            "children": [],
        },
        partial=True,
    )
    assert isinstance(user._id, str)
    assert isinstance(user.created_at, datetime.datetime)

    user = User(
        name="Aber",
        gender="m",
        age=18,
        tags=["a", "b"],
        wallet=Wallet(balance=100),
        children=[],
    )
    assert hasattr(user, "_id")
    assert isinstance(User.dump(user)["_id"], str)
    assert isinstance(user.created_at, datetime.datetime)


def test_recursion_field():
    user = User.load(
        {
            "name": "Aber",
            "gender": "m",
            "age": 18,
            "tags": ["a", "b"],
            "wallet": {"balance": 100},
            "children": [
                {
                    "name": "Yue",
                    "gender": "f",
                    "age": 18,
                    "tags": ["a", "b"],
                    "wallet": {"balance": 100},
                    "children": [],
                }
            ],
        }
    )
    assert isinstance(user.wallet, Wallet)
    assert user.wallet.balance == 100
    assert isinstance(user.children[0], User)
    assert user.children[0].name == "Yue"
    assert isinstance(user.children[0]._id, str)


def test_empty_field():
    user = User.load({}, partial=True)
    assert not hasattr(user, "name")
    assert hasattr(user, "_id")


def test_dict_field():
    user = User.load(
        {
            "name": "Aber",
            "gender": "m",
            "age": 18,
            "tags": ["a", "b"],
            "wallet": {"balance": 100},
            "children": [],
            "extra": {"a": "b"},
        }
    )
    assert user.extra == {"a": "b"}
    assert user.dump()["extra"] == {"a": "b"}


def test_datetime_field():
    user = User.load(dict(created_at=datetime.datetime.now()), partial=True)
    assert isinstance(user.created_at, datetime.datetime)


def test_literal_field():
    user = User.load(dict(gender="m"), partial=True)
    assert user.gender == "m"


def test_embedded_field():
    user = User.load({"wallet": {"balance": 100}}, partial=True)
    assert isinstance(user.wallet, Wallet)
    assert user.wallet.balance == 100


class UserWithRole(User):
    role: mongo.LiteralField[Literal["admin", "user"]]


UserWithRole.__lazy_init_fields__()


def test_three_level_inheritance():
    user = UserWithRole.load(dict(role="admin"), partial=True)
    assert isinstance(user._id, str)


class R0(mongo.Document):
    role: mongo.LiteralField[Literal["admin"]]
    t: mongo.StringField = mongo.StringField(default="r0")


class R1(mongo.Document):
    role: mongo.LiteralField[Literal["user"]]
    t: mongo.StringField = mongo.StringField(default="r1")


class U(mongo.Document):
    normal_type: mongo.UnionField[int | str]
    list_type: mongo.ListField[int | str]
    embedded_type: mongo.UnionField[R0 | R1]
    list_embedded_type: mongo.ListField[R0 | R1]


U.__lazy_init_fields__()


def test_union_field():
    u = U.load({"normal_type": 1}, partial=True)
    assert u.normal_type == 1
    u = U.load({"normal_type": "1"}, partial=True)
    assert u.normal_type == "1"

    u = U.load({"list_type": [1, "1"]}, partial=True)
    assert u.list_type == [1, "1"]

    u = U.load({"embedded_type": {"role": "admin"}}, partial=True)
    assert isinstance(u.embedded_type, R0)
    assert u.embedded_type.role == "admin"

    u = U.load({"embedded_type": {"role": "user"}}, partial=True)
    assert isinstance(u.embedded_type, R1)
    assert u.embedded_type.role == "user"

    u = U.load(
        {"list_embedded_type": [{"role": "admin"}, {"role": "user"}]}, partial=True
    )
    assert isinstance(u.list_embedded_type[0], R0)
    assert u.list_embedded_type[0].role == "admin"
    assert isinstance(u.list_embedded_type[1], R1)
    assert u.list_embedded_type[1].role == "user"
    assert u.dump() == {
        "list_embedded_type": [
            {"role": "admin", "t": "r0"},
            {"role": "user", "t": "r1"},
        ]
    }

    u = U(
        normal_type="normal",
        list_type=[1, "2"],
        embedded_type=R0(role="admin"),
        list_embedded_type=[R0(role="admin"), R1(role="user")],
    )
    assert u.to_mongo() == {
        "normal_type": "normal",
        "list_type": [1, "2"],
        "embedded_type": {"role": "admin", "t": "r0"},
        "list_embedded_type": [
            {"role": "admin", "t": "r0"},
            {"role": "user", "t": "r1"},
        ],
    }


class NotInitialized(mongo.Document):
    name: mongo.StringField


def test_not_initialized():
    with pytest.raises(
        AttributeError,
        match="Please initialize the Document NotInitialized before using it.",
    ):
        NotInitialized.name


def test_type_to_field_edge_cases():
    # 测试所有基本类型
    assert type_to_field(str).__class__.__name__ == "StringField"
    assert type_to_field(int).__class__.__name__ == "IntegerField"
    assert type_to_field(float).__class__.__name__ == "FloatField"
    assert type_to_field(bool).__class__.__name__ == "BooleanField"
    assert type_to_field(datetime.datetime).__class__.__name__ == "DateTimeField"
    assert type_to_field(Decimal).__class__.__name__ == "DecimalField"
    assert type_to_field(ObjectId).__class__.__name__ == "ObjectIdField"

    # 测试无效类型
    with pytest.raises(ValueError) as exc:
        type_to_field(complex)
    assert "Cannot convert type" in str(exc.value)


class Mode(str, enum.Enum):
    UNLIMITED = "unlimited"
    LIMITED = "limited"


class Settings(MongoDocument):
    language: mongo.LiteralField[Literal["en", "zh"]]
    mode: mongo.EnumField[Mode]


def test_pydantic_schema():
    from pydantic import BaseModel

    class UserEntity(BaseModel):
        settings: Settings

    user = UserEntity.model_validate(
        {"settings": {"language": "en", "mode": "unlimited"}}
    )
    assert user.settings.language == "en"
    assert user.settings.mode == Mode.UNLIMITED

    schema = UserEntity.model_json_schema()
    assert schema == {
        "$defs": {
            "Mode": {
                "enum": ["unlimited", "limited"],
                "title": "Mode",
                "type": "string",
            }
        },
        "properties": {
            "settings": {
                "properties": {
                    "_id": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "title": "Id",
                    },
                    "language": {
                        "anyOf": [
                            {"enum": ["en", "zh"], "type": "string"},
                            {"type": "null"},
                        ],
                        "title": "Language",
                    },
                    "mode": {"anyOf": [{"$ref": "#/$defs/Mode"}, {"type": "null"}]},
                },
                "required": ["language", "mode"],
                "title": "Settings",
                "type": "object",
            }
        },
        "required": ["settings"],
        "title": "UserEntity",
        "type": "object",
    }



from __future__ import annotations

import dataclasses
import inspect
from functools import reduce
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Mapping,
    Optional,
    Sequence,
    get_args,
    get_origin,
)

from marshmallow import Schema
from pymongo import IndexModel
from typing_extensions import Self, dataclass_transform

from typedmongo.exceptions import DocumentDefineError
from typedmongo.marshamallow import MarshamallowObjectId

from .client import Manager
from .fields import Field, ListField, ObjectIdField, type_to_field

if TYPE_CHECKING:
    from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
    from pydantic.json_schema import JsonSchemaValue
    from pydantic_core import core_schema


def snake_case(name: str) -> str:
    """
    convert "SomeWords" to "some_words"
    """
    return "".join(
        "_" + char.lower() if char.isupper() and i > 0 else char.lower()
        for i, char in enumerate(name)
    )


@dataclasses.dataclass
class Index:
    keys: (
        Field
        | Sequence[tuple[Field, int | str | Mapping[str, Any]]]
        | Mapping[Field, Any]
        | Mapping[str, Any]
    )

    name: Optional[str] = None
    unique: bool = False
    background: bool = False
    sparse: bool = False
    expireAfterSeconds: Optional[int] = None
    partialFilterExpression: Optional[dict[str, Any]] = None

    def to_index_model(self) -> IndexModel:
        if isinstance(self.keys, Field):
            keys = self.keys.field_name
        elif isinstance(self.keys, Mapping):
            keys = {
                field.field_name if isinstance(field, Field) else field: value
                for field, value in self.keys.items()
            }
        else:
            keys = [(field.field_name, value) for field, value in self.keys]

        # Don't use asdict, because it will raise RecursionError
        kwargs = {
            key: value
            for key, value in self.__dict__.items()
            if value is not None and key != "keys"
        }

        return IndexModel(keys=keys, **kwargs)


@dataclass_transform(kw_only_default=True)
class DocumentMetaClass(type):
    if TYPE_CHECKING:
        __abstract__: bool
        __database__: Any
        __collection_name__: str
        __collection__: Any
        __pfields__: dict[str, Field]
        __sfields__: dict[str, Field]
        __fields_loaded__: bool
        __fields__: dict[str, Field]
        __create_marshamallow_schema__: Callable[[str, dict[str, Field]], Schema]
        __schema__: Schema

    def __new__(cls, name, bases, namespace):
        if "_" in name:  # check error name. e.g. Status_Info
            raise DocumentDefineError(
                "Document class name cannot have '_': {name}".format(name=name)
            )
        if not name[0].isupper():  # check error name. e.g. statusInfo
            raise DocumentDefineError(
                "Document class name must be upper letter in start: {name}".format(
                    name=name
                )
            )
        namespace.setdefault("__collection_name__", snake_case(name))
        namespace.setdefault("__abstract__", False)

        if "__fields__" in namespace:  # Clear __fields__ to avoid conflict
            del namespace["__fields__"]
        namespace["__fields_loaded__"] = False

        namespace["__sfields__"] = {
            name: value for name, value in namespace.items() if isinstance(value, Field)
        }
        # merge bases' `__fields__` to `__pfields__`
        namespace["__pfields__"] = reduce(
            lambda _initial, _item: {**_item, **_initial},
            reversed(
                [
                    base.__lazy_init_fields__()
                    for base in bases
                    if isinstance(base, DocumentMetaClass)
                ]
            ),
            {},
        )

        if "__create_marshamallow_schema__" not in namespace:
            for base in bases:
                create_schema = base.__create_marshamallow_schema__
            namespace["__create_marshamallow_schema__"] = create_schema

        return super().__new__(cls, name, bases, namespace)

    def __lazy_init_fields__(cls) -> dict[str, Field]:
        if cls.__fields_loaded__:
            return cls.__fields__

        fields = {
            **{
                name: value()
                for name, value in inspect.get_annotations(cls, eval_str=True).items()
                if isinstance(value, type) and issubclass(value, Field)
            },
            **{
                name: origin_class(type_to_field(get_args(value)[0]))
                if issubclass(origin_class, ListField)
                else origin_class(*get_args(value))
                for name, value in inspect.get_annotations(cls, eval_str=True).items()
                if (origin_class := get_origin(value))
                and isinstance(origin_class, type)
                and issubclass(origin_class, Field)
            },
            **cls.__sfields__,
        }
        cls.__fields__ = {**cls.__pfields__, **fields}
        cls.__fields_loaded__ = True

        for name, field in fields.items():
            setattr(cls, name, field)
            field.__set_name__(cls, name)

        cls.__schema__ = cls.__create_marshamallow_schema__(
            cls.__name__, cls.__fields__
        )
        return cls.__fields__

    def __setattr__(cls, name, value):
        if name == "__abstract__":
            raise AttributeError(
                "Can't modify the `__abstract__` attribute dynamically."
            )
        return super().__setattr__(name, value)

    def __getattribute__(self, name: str) -> Any:
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if not self.__fields_loaded__:
                message = "Please initialize the Document {class_name} before using it.".format(
                    class_name=self.__name__
                )
                raise AttributeError(message, name=name, obj=self)
            raise


class Document(metaclass=DocumentMetaClass):
    __abstract__ = True

    objects = Manager()

    def __init__(self, **kwargs):
        if self.__abstract__:
            raise RuntimeError(
                "The class {} cannot be instantiated, because it's __abstract__ is True.".format(
                    self.__class__.__name__
                )
            )

        for name, field in self.__class__.__fields__.items():
            if name in kwargs:
                value = kwargs.pop(name)
            else:
                if field.default is None:
                    continue
                default_value = field.default
                if callable(default_value):
                    value = default_value()
                else:
                    value = default_value
            setattr(self, name, value)

        if kwargs:
            raise TypeError(
                "{class_name}() got unexpected keyword arguments '{unexpected}'".format(
                    class_name=self.__class__.__name__,
                    unexpected="', '".join(kwargs.keys()),
                )
            )

    def __repr__(self) -> str:
        return "{class_name}({fields})".format(
            class_name=self.__class__.__name__,
            fields=", ".join(
                f"{name}={repr(self.__dict__[name])}"
                for name, _ in self.__fields__.items()
                if name in self.__dict__
            ),
        )

    def __eq__(self, value: object) -> bool:
        return isinstance(value, self.__class__) and {
            key: self.__dict__.get(key, None) for key in self.__fields__.keys()
        } == {key: value.__dict__.get(key, None) for key in value.__fields__.keys()}

    @staticmethod
    def __create_marshamallow_schema__(name: str, fields: dict[str, Field]) -> Schema:
        schema_class = Schema.from_dict(
            {name: field.marshamallow for name, field in fields.items()}, name=name
        )
        return schema_class(unknown="exclude")

    @classmethod
    def load(
        cls: type[Self], data: Mapping[str, Any], *, partial: bool = False
    ) -> Self:
        """
        Load data from dict to instance, and validate the data.

        :param data: The data to load.
        :param partial: If True, allow partial data to load.
        """
        validated: dict[str, Any] = cls.__schema__.load(data=data, partial=partial)  # type: ignore
        loaded = {
            key: getattr(cls.__fields__[key], "load")(value, partial=partial)
            for key, value in validated.items()
        }
        return cls(**loaded)

    def dump(self: Self) -> dict[str, Any]:
        """
        Dump the instance to jsonable dict.
        """
        return self.__schema__.dump(self)  # type: ignore

    @classmethod
    def indexes(cls) -> list[Index]:
        return []

    def to_mongo(self) -> dict[str, Any]:
        """
        Dump the instance to dict for mongo.
        """
        return {
            key: getattr(self.__fields__[key], "to_mongo")(value)
            for key, value in self.__dict__.items()
        }

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """
        Make the Document class can be validated by Pydantic
        """
        if cls.__abstract__:
            raise TypeError(
                f"Cannot use abstract class {cls.__name__} as a Pydantic type"
            )

        # Ensure fields are initialized
        cls.__lazy_init_fields__()

        from pydantic_core import core_schema

        instance_schema = core_schema.is_instance_schema(cls)
        dict_schema = handler.generate_schema(dict[str, Any])

        return core_schema.union_schema(
            [
                instance_schema,
                core_schema.no_info_after_validator_function(cls.load, dict_schema),
            ]
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        """
        Make the Document class can be understood by Pydantic
        """
        from pydantic import ConfigDict, Field, create_model

        fields = {}
        for name, field in cls.__fields__.items():
            is_optional = field.allow_none
            has_default = field.default is not None

            alias = name if name.startswith("_") else None

            if is_optional:
                pydantic_type = Optional[field.field_type]
            else:
                pydantic_type = field.field_type

            if has_default:
                if callable(field.default):
                    field_info = Field(default_factory=field.default, alias=alias)
                else:
                    field_info = Field(default=field.default, alias=alias)
            else:
                field_info = Field(..., alias=alias)

            fields[name.lstrip("_")] = (pydantic_type, field_info)

        pydantic_model = create_model(
            cls.__name__, __config__=ConfigDict(defer_build=True), **fields
        )
        return handler(pydantic_model.__pydantic_core_schema__)


class MongoDocument(Document):
    __abstract__ = True

    _id: ObjectIdField = ObjectIdField(
        marshamallow=MarshamallowObjectId(required=False)
    )
